import subprocess
from pathlib import Path
import click

# FIX need to check for rsync and ssh

#  ──────────────────────────────────────────────────────────────────────────

class Cache():

    def __init__(self, config, initialize):
        self.config = config['cache']
        self.path = Path(self.config['path'])

        if initialize:
            self.init()


    def init(self):
        command = ['mkdir -p', self.path / Path('annotated') ]
        subprocess.run(command)


    def write(self, local_file_path, remote_file_path, force=False):

        command = ['rsync']
        
        if force:
            command += ['-I']

        command += [Path(local_file_path).resolve(), self.host + ':' + str(self.path / Path(remote_file_path))]
        subprocess.run(command)


    def delete(self, remote_file_path):
        command = ['ssh', self.host, 'rm', self.path / Path(remote_file_path)]
        click.echo(command)
        subprocess.run(command)


    def move(self, old_remote_file_path, new_remote_file_path):
        command = ['ssh', self.host, 'mv', self.path / Path(old_remote_file_path), self.path / Path(new_remote_file_path)]
        click.echo(command)
        subprocess.run(command)



import tomllib
from pathlib import Path

#  ──────────────────────────────────────────────────────────────────────────

# default configuration
config = {'index': {'path': '~/.bibtheque/index.json'},
          'database': {'host': '0.0.0.0',
                       'host_port': 2340,
                       'path': '~/.bibtheque_db/',
                       'cache': '~/.cache/bibtheque',
                       'cache_port': 2340,
                       },
          }

# default config path
config_path = Path("~/.config/bibtheque.toml").expanduser()

if config_path.exists():
    with open(config_path, 'rb') as file:
        tmp = tomllib.load(file)

    # set configuration from toml
    for key in tmp.keys():
        config[key] = tmp[key]



from pypdf import PdfReader
import click
import json
import re
from pathlib import Path

#  ──────────────────────────────────────────────────────────────────────────
# local imports

import bibtheque.fields as fields
import bibtheque.document as document
import bibtheque.strings as strings

#  ──────────────────────────────────────────────────────────────────────────

class Index():

    def __init__(self, config, initialize=False):
        self.config = config['index']
        self.path = Path(self.config['path']).expanduser()

        # handling initialization
        if initialize:
            self.init()
        else:
            self.read()


    def init(self):
        """Initializes the index.json file"""
        if not self.path.exists():
            self.path.parent.mkdir(parents=True)

        with click.open_file(self.path, 'w') as file:
            json.dump({}, file)


    def read(self):
        """Reads the index.json file"""
        with click.open_file(self.path, 'r') as file:
            self.index = json.load(file)


    def write(self):
        """Write changes to index file"""
        with click.open_file(self.path, 'w') as file:
            json.dump(self.index, file)


    def exists(self, doc_id):
        if doc_id in self.index:
            return True
        else:
            return False


    def insert(self, doc_id, doc, force=False):
        """Inserts document into the index"""

        # looking for duplicates
        ignore_common_bib_fields = list(set(fields.all_fields) - set(['publisher', 'url', 'type', 'month', 'year', 'number', 'volume']))
        duplicates = self.find_duplicates(doc, bib_fields=ignore_common_bib_fields)

        if force or not duplicates:
            self.index[doc_id] = doc
        else:
            doc_str = json.dumps(doc, indent=4)
            json_str = json.dumps(duplicates, indent=4)
            raise click.ClickException(click.style('Duplicates found:\n', bold=True) + click.style('Submitted:\n', bold=True) + click.style(doc_str) + click.style('\nDuplicates:\n', bold=True) + click.style(json_str))


    def modify(self, doc_id):
        if self.exists(doc_id):
            old = self.index[doc_id]

            tmp = json.dumps(self.index[doc_id])
            tmp = click.edit(tmp, extension='.json')

            if tmp: # if changes are made, return the differences or return Nones
                self.index[doc_id] = json.loads(tmp) # index modified here

                # getting differences
                tmp = json.loads(tmp)
                diff = {}
                for attr in tmp:
                    if attr == 'bib':

                        diff[attr] = {}
                        for field in tmp[attr]:
                            if field not in old[attr] or tmp[attr][field] != old[attr][field]:
                                diff[attr][field] = tmp[attr][field]

                        if not diff[attr]:
                            del diff[attr]

                    elif attr == 'tags':
                        diff[attr] = []
                        for tag in tmp[attr]:
                            if tag not in old[attr]:
                                diff[attr].append(tag)

                        if not diff[attr]:
                            del diff[attr]

                    else:
                        if attr not in old or tmp[attr] != old[attr]:
                            diff[attr] = tmp[attr]

                if diff:
                    return old, diff
                else:
                    return None, None
            else:
                return None, None

        else:
            raise click.ClickException('Doc not found')


    def delete(self, doc_id):
        """Delete doc from the index"""
        if self.exists(doc_id):
            del self.index[doc_id]


    def search(self, regex, lower=True):
        """Search through index with regex"""

        def wrap_search(regex, to_be_searched, lower):

            if lower:
                regex = regex.lower()
                to_be_searched = to_be_searched.lower()

            return re.search(regex, to_be_searched)


        def wrap_finditer(regex, to_be_searched, lower):

            if lower:
                regex = regex.lower()
                to_be_searched = to_be_searched.lower()

            return re.finditer(regex, to_be_searched)


        hits = {}

        for doc_id in self.index:

            hit = {}

            for attr in self.index[doc_id]:

                # search through
                if attr == 'bib':

                    hit[attr] = {}

                    for field in self.index[doc_id][attr]:
                        if wrap_search(regex, self.index[doc_id][attr][field], lower):
                            hit[attr][field] = wrap_finditer(regex, self.index[doc_id][attr][field], lower)

                    if not hit[attr]:
                        del hit[attr]

                elif attr == 'tags':

                    hit[attr] = {}

                    for tag in self.index[doc_id][attr]:
                        if wrap_search(regex, tag, lower):
                            hit[attr][tag] = wrap_finditer(regex, tag, lower)

                    if not hit[attr]:
                        del hit[attr]

                else:
                    if wrap_search(regex, self.index[doc_id][attr], lower):
                        hit[attr] = wrap_finditer(regex, self.index[doc_id][attr], lower)

            if hit:
                hits[doc_id] = hit
        
        return hits


    def find_duplicates(self, doc, search_docs=None, doc_fields=None, bib_fields=None, duplicate_threshold=None, length_threshold=None):
        """Searches docs by field for duplicate fields"""

        def _find(self, doc, search_docs, doc_fields, bib_fields, duplicate_threshold, length_threshold):
            """Searches docs by field for duplicate fields: ACTUAL LOGIC"""

            duplicates = {}
            for doc_id in search_docs:

                duplicate_doc = {}

                # search doc_fields
                for doc_field in doc_fields:

                    if doc_field in doc and doc_field in search_docs[doc_id]:

                        # search by bib_fields
                        if doc_field == 'bib':

                            duplicate_doc[doc_field] = {}

                            for bib_field in bib_fields:
                                if bib_field in doc[doc_field] and bib_field in search_docs[doc_id][doc_field]:
                                    if 0 < len(doc[doc_field][bib_field]) < length_threshold and 0< len(search_docs[doc_id][doc_field][bib_field]) < length_threshold:
                                        if strings.matching(doc[doc_field][bib_field], search_docs[doc_id][doc_field][bib_field], duplicate_threshold):
                                            duplicate_doc[doc_field][bib_field] = search_docs[doc_id][doc_field][bib_field]

                            # removing if no duplicates in bib_fields are found
                            if not duplicate_doc[doc_field]:
                                del duplicate_doc[doc_field]

                        # ignore tags
                        elif doc_field == 'tags':
                            pass

                        # search everything else normally

                        elif 0 < len(doc[doc_field]) < length_threshold and 0 < len(search_docs[doc_id][doc_field]) < length_threshold:
                            if strings.matching(doc[doc_field], search_docs[doc_id][doc_field], duplicate_threshold):
                                duplicate_doc[doc_field] = search_docs[doc_id][doc_field]

                if duplicate_doc:
                    duplicates[doc_id] = duplicate_doc

            return duplicates


        def _search_fields(search_fields, all_fields):
            """Setting search fields"""

            if not search_fields:
                search_fields = all_fields
            else:
                for field in search_fields:
                    if field not in all_fields:
                        click.echo(field, 'not an allowable field') # FIX
                        raise click.Abort

            return search_fields


        # setting search_docs to all docs by default
        if not search_docs:
            search_docs = self.index

        # set doc_fields to search through
        doc_fields = _search_fields(doc_fields, document.doc_fields)

        # sets bib_fields to all if none are specified
        bib_fields = _search_fields(bib_fields, fields.all_fields)

        # setting default threshold from config
        if not duplicate_threshold:
            duplicate_threshold = self.config['duplicate_threshold']

        # setting length threshold from config
        if not length_threshold:
            length_threshold = self.config['length_threshold']

        duplicates = _find(self, doc, search_docs, doc_fields, bib_fields, duplicate_threshold, length_threshold)

        return duplicates


    def find_all_duplicates(self):
        """Find all duplicates in the index"""

        all_duplicates = {}

        for pivot_doc_id in self.index:

            search_docs = {}

            for searchable_doc_id in self.index:
                if pivot_doc_id != searchable_doc_id:
                    search_docs[searchable_doc_id] = self.index[searchable_doc_id]

            all_duplicates[pivot_doc_id] = self.find_duplicates(self.index[pivot_doc_id], search_docs=search_docs)

        return all_duplicates


    def read_pdf(self, pdf_path):
        """Extracts text from PDF"""

        if Path(pdf_path).suffix == '.pdf':

            # extracting text
            readpdf = PdfReader(pdf_path)
            pdf_text = ''
            for page in readpdf.pages:
                pdf_text += page.extract_text()

            return pdf_text

        else:
            click.echo('Not a pdf')
            return None


    def file_exists(self, doc_id, other=False):
        """other is either ['annotated', 'supplemental']"""
        if other:
            if other in ['annotated', 'supplemental']:
                path_name = other + '_path'
            else:
                click.echo(other + ' not an allowable field')
                return None

        else:
            path_name = 'file_path'

        if self.index[doc_id][path_name]:
            return True
        else:
            return False



# ══════════════════════════════════════════════════════════════════════════
# test URLs, DOIs, and ISBNs

# ──────────────────────────────────────────────────────────────────────────
# test URLs

book_url = "https://www.sciencedirect.com/book/9780123848666/an-introduction-to-dynamic-meteorology"
book_isbn = "978-0-12-384866-6"
book_doi = "https://doi.org/10.1016/C2009-0-63394-8"

#  ──────────────────────────────────────────────────────────────────────────
# DOI

doi_url = "https://doi.org/10.1002/2016JC011857"
doi_str = "10.1002/2016JC011857"
doi_bib = "@article{Houpert_2016, title={Observations of open-ocean deep convection in the northwestern Mediterranean Sea: Seasonal and interannual variability of mixing and deep water masses for the 2007-2013 Period: DEEP CONVECTION OBS. NWMED 2007-2013}, volume={121}, ISSN={2169-9275}, url={http://dx.doi.org/10.1002/2016JC011857}, DOI={10.1002/2016jc011857}, number={11}, journal={Journal of Geophysical Research: Oceans}, publisher={American Geophysical Union (AGU)}, author={Houpert, L. and Durrieu de Madron, X. and Testor, P. and Bosse, A. and D’Ortenzio, F. and Bouin, M. N. and Dausse, D. and Le Goff, H. and Kunesch, S. and Labaste, M. and Coppola, L. and Mortier, L. and Raimbault, P.}, year={2016}, month=nov, pages={8139–8171} }"

#  ──────────────────────────────────────────────────────────────────────────
# ISBN

isbn10 = "0-486-60061-0"
isbn13 = "9780486600611"
isbn_bib = """@book{9780486600611,
     title = {Fundamentals Of Astrodynamics},
    author = {Roger R. Bate and Donald D. Mueller and Jerry E. White},
      isbn = {9780486600611},
      year = {1971},
 publisher = {Courier Corporation}
}"""



import click
import sys
import importlib.metadata
import json
import re
from pathlib import Path
import subprocess
from shutil import which
import http.server
import socketserver

#  ──────────────────────────────────────────────────────────────────────────
# local imports

import bibtheque.document as document
import bibtheque.index as index
import bibtheque.database as database
from bibtheque.config import config
import bibtheque.fields as fields


#  ──────────────────────────────────────────────────────────────────────────
# global variables

bibtheque_version = importlib.metadata.version('bibtheque')
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


#  ──────────────────────────────────────────────────────────────────────────
# global functions

def handle_tags(tags):
    """Handle comma delimited tag string"""
    tag_list = tags.split(',')
    tags = []
    for tag in tag_list:
        tags.append(tag.strip())
    return sorted(set(tags))


#  ──────────────────────────────────────────────────────────────────────────
# global options

file_option = click.option('-f', '--file', 'file', default="", help='PDF file path', type=str)
annotated_option = click.option('-a', '--annotated', 'annotated', default="", help='Annotated PDF file path', type=str)
supplemental_option = click.option('-si', '--supplemental', 'supplemental', default="", help='Supplemental PDF file path', type=str)
tags_option = click.option('-t', '--tags', 'tags', default='', help="Comma delimited tags; e.g. 'tag0, tag1, tag2'")
force_option = click.option('--force', is_flag=True, default=False, show_default=True, help='Force command', type=bool)
dry_run_option = click.option('--dry-run', 'dry_run', is_flag=True, default=False, show_default=True, help='Dry run of command')


#  ──────────────────────────────────────────────────────────────────────────
# global arguments

doc_id_arg = click.argument('doc_id', type=str)


#  ──────────────────────────────────────────────────────────────────────────
# base command

@click.group(invoke_without_command=True, context_settings=CONTEXT_SETTINGS)
@click.version_option(bibtheque_version)
@click.option('--config', type=dict, default=config, help='Configuration')
@click.option('--debug', type=int, default=1, help='Debug traceback level; higher prints more: e.g. 3 prints 3 levels of traceback; a negative value prints all traceback')
@click.option('--init', is_flag=True, default=False, help='Init')
@click.pass_context
def bibtheque(ctx, config, debug, init):
    """A tool to manage bibliographic items, particularily for research.
    """

    # check if rsync and ssh are available
    if which('rsync') and which('ssh'):
        pass
    else:
        raise click.ClickException('rsync and ssh not found')

    # setting debug traceback; value of None prints all
    if debug < 0:
        debug = None

    sys.tracebacklimit = debug

    if init and click.confirm('Initializing an existing Bibtheque index will delete it. Are you sure you want to continue?'):
        pass
    else:
        init = False

    # getting configuration
    ctx.obj = {'index': index.Index(config, initialize=init), 'database': database.Database(config, bibtheque_version, initialize=init), 'config': config}

    # initializing git
    if init:
        ctx.invoke(git, git_commands=['init'])
        ctx.invoke(git, git_commands=['add', ctx.obj['index'].path])
        ctx.invoke(git, git_commands=['commit', '-m', '"Initialization"'])


#  ──────────────────────────────────────────────────────────────────────────
# Insert

def build_doc(ctx, doc_id, doc_bib_dict, file, annotated, supplemental, tags, notes, synopsis):
    """Insert command"""

    # setting up document dictionary

    # adding pdf file text and path
    if file:
        # check if file exists
        with click.open_file(file) as _:
            pass

        # file_text = ctx.obj['index'].read_pdf(file) # DISABLING FOR NOW; NEED TO MAKE ADDITIONAL COMMAND TO EXTRACT PDF TEXT AND PLACE IN REPOSITORY AS SEPARATE FILE
        file_text = ""
        file_path = doc_id + '.pdf'
    else:
        file_text = ""
        file_path = ""

    # adding annotated file path
    if annotated:
        # check if annotated file exists
        with click.open_file(annotated) as _:
            pass

        annotated_path = 'annotated/' + doc_id + '.pdf'
    else:
        annotated_path = ""

    # adding supplemental file path
    if supplemental:
        # check if supplemental file exists
        with click.open_file(supplemental) as _:
            pass

        supplemental_path = 'supplemental/' + doc_id + '.pdf'
    else:
        supplemental_path = ""

    # listifying tags
    if tags:
        tags = handle_tags(tags)
    else:
        tags = []

    # build doc
    doc = document.build_doc(doc_bib_dict, file_text=file_text, file_path=file_path, annotated_path=annotated_path, supplemental_path=supplemental_path, tags=tags, notes=notes, synopsis=synopsis)

    return doc, file_path, annotated_path, supplemental_path


def dry_run_exit(doc_id, doc, dry_run):
    """Dry run exit point"""

    if dry_run:
        dry_run_doc = json.dumps(doc, indent=4)
        click.echo(click.style(doc_id, bold=True) + '\n' + dry_run_doc)

        if click.confirm('Do you wish to insert?'):
            pass
        else:
            raise click.Abort()


def soft_insert_doc(ctx, doc_id, doc, force, dry_run):
    """Soft DOC insert"""

    dry_run_exit(doc_id, doc, dry_run)

    # inserting into index
    ctx.obj['index'].insert(doc_id, doc, force=force)


def insert_doc(ctx, doc_id, doc, file, file_path, annotated, annotated_path, supplemental, supplemental_path, force):
    """Inserting DOC"""

    # writing to index and database
    if (file or annotated or supplemental) and ctx.obj['database'].available(): # checking connectivity

        # adding file to file repository
        if file:
            ctx.obj['database'].write(file, file_path, force=force)

        if annotated:
            ctx.obj['database'].write(annotated, annotated_path, force=force)

        if supplemental:
            ctx.obj['database'].write(supplemental, supplemental_path, force=force)

    else:
        raise click.ClickException('Cannot connect to database')


    # complete write when all else passes
    ctx.obj['index'].write()

    # git adding insert
    ctx.invoke(git, git_commands=['add', ctx.obj['index'].path])
    ctx.invoke(git, git_commands=['commit', '-m', '"Inserted ' + doc_id + '"'])

    # confirmation
    click.echo(click.style('Inserted: ', bold=True) + doc_id)


# insert command
@click.command()
@file_option
@annotated_option
@supplemental_option
@tags_option
@click.option('-n', '--notes', 'notes', default='', help='Additional notes', type=str)
@click.option('-s', '--synopsis', 'synopsis', default='', help='Synopsis of the work', type=str)
@force_option
@dry_run_option
@click.argument('standard', type=str)
@click.pass_context
def insert(ctx, file, annotated, supplemental, tags, notes, synopsis, force, dry_run, standard):
    """Insert a document into the database."""

    # build doc dict and id
    doc_bib_dict = document.determine_standard(standard)
    doc_id = document.generate_id(standard)

    # regenerating if index already exists
    while ctx.obj['index'].exists(doc_id):
        doc_id = document.generate_id(standard)

    doc, file_path, annotated_path, supplemental_path = build_doc(ctx, doc_id, doc_bib_dict, file, annotated, supplemental, tags, notes, synopsis)
    soft_insert_doc(ctx, doc_id, doc, force, dry_run)
    insert_doc(ctx, doc_id, doc, file, file_path, annotated, annotated_path, supplemental, supplemental_path, force)

bibtheque.add_command(insert)


# manual
@click.command()
@file_option
@annotated_option
@supplemental_option
@tags_option
@click.option('-n', '--notes', 'notes', default='', help='Additional notes', type=str)
@click.option('-s', '--synopsis', 'synopsis', default='', help='Synopsis of the work', type=str)
@force_option
@dry_run_option
@click.argument('type', type=str)
@click.pass_context
def manual(ctx, file, annotated, supplemental, tags, notes, synopsis, force, dry_run, type):
    """Insert a document into the database."""

    # build doc dict and id
    doc_bib_dict = {}
    for field in fields.required[type]:
        if field == "type":
            doc_bib_dict[field] = type
        else:
            doc_bib_dict[field] = ""

    doc_id = document.generate_id(type)

    # regenerating if index already exists
    while ctx.obj['index'].exists(doc_id):
        doc_id = document.generate_id(type)
    

    doc, file_path, annotated_path, supplemental_path = build_doc(ctx, doc_id, doc_bib_dict, file, annotated, supplemental, tags, notes, synopsis)

    # inserting into index
    ctx.obj['index'].insert(doc_id, doc, force=force)
    _, _ = ctx.obj['index'].modify(doc_id)

    dry_run_exit(doc_id, doc, dry_run)
    insert_doc(ctx, doc_id, doc, file, file_path, annotated, annotated_path, supplemental, supplemental_path, force)

bibtheque.add_command(manual)


#  ──────────────────────────────────────────────────────────────────────────
# Modify

@click.command()
@file_option
@annotated_option
@supplemental_option
@tags_option
@force_option
@doc_id_arg
@click.pass_context
def modify(ctx, file, annotated, supplemental, tags, force, doc_id):
    """Opens the default text editor to edit the document with the given DOC_ID."""

    # adding file
    if file:
        attr = 'file_path'
        overwrite = False
        if attr in ctx.obj['index'].index[doc_id]:
            if force or click.confirm('Are you sure you want overwrite the current file path?'):
                overwrite = True
        else:
            overwrite = True

        if overwrite:
            file_text = ctx.obj['index'].read_pdf(file)
            file_path = doc_id + '.pdf'
            ctx.obj['index'].index[doc_id][attr] = file_path
            ctx.obj['index'].write()
            ctx.obj['database'].write(file, file_path, force=force)

        return

    # adding annotated file
    if annotated:
        attr = 'annotated_path'
        overwrite = False
        if attr in ctx.obj['index'].index[doc_id]:
            if force or click.confirm('Are you sure you want overwrite the current annotated file path?'):
                overwrite = True
        else:
            overwrite = True

        if overwrite:
            annotated_path = 'annotated/' + doc_id + '.pdf'
            ctx.obj['index'].index[doc_id][attr] = annotated_path
            ctx.obj['index'].write()
            ctx.obj['database'].write(annotated, annotated_path, force=force)

        return

    # adding supplemental file
    if supplemental:
        attr = 'supplemental_path'
        overwrite = False
        if attr in ctx.obj['index'].index[doc_id]:
            if force or click.confirm('Are you sure you want overwrite the current supplemental file path?'):
                overwrite = True
        else:
            overwrite = True

        if overwrite:
            supplemental_path = 'supplemental/' + doc_id + '.pdf'
            ctx.obj['index'].index[doc_id][attr] = supplemental_path
            ctx.obj['index'].write()
            ctx.obj['database'].write(supplemental, supplemental_path, force=force)

        return

    # adding tags
    if tags:
        attr = 'tags'
        old_tags = ctx.obj['index'].index[doc_id][attr]
        old_tags = ','.join(old_tags)
        tags = handle_tags(tags + ',' + old_tags)
        ctx.obj['index'].index[doc_id][attr] = tags
        ctx.obj['index'].write()

        return

    old, diff = ctx.obj['index'].modify(doc_id)
    ctx.obj['index'].write()
    
    # handling changes in file and annotated file paths without adding a new file
    if diff:
        for attr in ['file_path', 'annotated_path']:
            if attr in diff:
                if force or click.confirm('Are you sure you want to change the file or annotated file path?'):
                    ctx.obj['database'].move(old[attr], diff[attr])

    # git adding delete
    ctx.invoke(git, git_commands=['add', ctx.obj['index'].path])
    ctx.invoke(git, git_commands=['commit', '-m', '"Modified ' + doc_id + '"'])

    # confirmation
    click.echo(click.style('Modified: ', bold=True) + doc_id)

bibtheque.add_command(modify)


#  ──────────────────────────────────────────────────────────────────────────
# Delete

@click.command()
@force_option
@doc_id_arg
@click.pass_context
def delete(ctx, force, doc_id):
    """Deletes the document with the given DOC_ID."""

    if click.confirm('Are you sure you want to delete this document?'):

        # deleting file
        if ctx.obj['index'].file_exists(doc_id):
            ctx.obj['database'].delete(ctx.obj['index'].index[doc_id]['file_path'])

        # deleting annotated file
        if ctx.obj['index'].file_exists(doc_id, other='annotated'):
            ctx.obj['database'].delete(ctx.obj['index'].index[doc_id]['annotated_path'])

        # deleting supplemental file
        if ctx.obj['index'].file_exists(doc_id, other='supplemental'):
            ctx.obj['database'].delete(ctx.obj['index'].index[doc_id]['annotated_path'])
        
        # deleting index entry; must be deleted last
        ctx.obj['index'].delete(doc_id)
        ctx.obj['index'].write()

        # git adding delete
        ctx.invoke(git, git_commands=['add', ctx.obj['index'].path])
        ctx.invoke(git, git_commands=['commit', '-m', '"Deleted ' + doc_id + '"'])

        # confirmation
        click.echo(click.style('Deleted: ', bold=True) + doc_id)

bibtheque.add_command(delete)


#  ──────────────────────────────────────────────────────────────────────────
# Search

# fields? need to add fields to only search through
@click.command()
@click.option('-o', '--output', 'output', default='counts', type=click.Choice(['minimal', 'counts', 'json', 'pretty']), help='Option that changes the output format.')
@click.option('-p', '--pager', 'pager', default=False, is_flag=True, help='Enters pager mode to page through results.')
@click.option('-u', '--url', 'url', default=False, is_flag=True, help='Output result as URL.')
@click.option('-b', '--bib', '--bibtex', 'bib', default=False, is_flag=True, help='Outputs documents found from the search in BiBTeX string format.')
@click.argument('regex', type=str, default="") # add bibtex output and handle tags by specifying tag?
@click.pass_context
def search(ctx, output, pager, url, bib, regex):
    """Searches the database with REGEX_TEXT.
    """

    def print_results(hits, output='simple', pager=False, host_url=None):
        """
        'minimal', 'counts', 'json', 'pretty'
        """

        results_json = {}

        for doc_id in hits:

            if output == 'minimal':
                click.echo(doc_id)
                continue

            counts = [] # for counts output
            result_json = {} # for json output
            pretty = [] # for pretty output

            for attr in hits[doc_id]:
                if attr == 'bib':
                    result_json[attr] = {}

                    for field in hits[doc_id][attr]:

                        tmp_pretty = ''
                        for i, match in enumerate(hits[doc_id][attr][field]):
                            result_json[attr][field] = match.string # just overwriting the same string

                            # pretty handling
                            match_span = match.span()
                            match_str = match.string

                            if i == 0:
                                prefix = match_str[0:match_span[0]]
                                tmp_pretty += prefix + click.style(match_str[match_span[0]:match_span[1]], bold=True)
                                prior_match_span = match_span
                            else:
                                prefix = match_str[prior_match_span[1]:match_span[0]]
                                tmp_pretty += prefix + click.style(match_str[match_span[0]:match_span[1]], bold=True)
                                prior_match_span = match_span

                        tmp_pretty += match_str[match_span[1]:]
                        pretty.append(attr + ': ' + field + ': ' + tmp_pretty)
                        counts.append(attr + ': ' + field + ': ' + str(i+1))

                elif attr == 'tags':
                    result_json[attr] = []

                    for tag in hits[doc_id][attr]:
                        result_json[attr].append(tag)

                        tmp_pretty = ''
                        for i, match in enumerate(hits[doc_id][attr][tag]):

                            # pretty handling
                            match_span = match.span()
                            match_str = match.string

                            if i == 0:
                                prefix = match_str[0:match_span[0]]
                                tmp_pretty += prefix + click.style(match_str[match_span[0]:match_span[1]], bold=True)
                                prior_match_span = match_span
                            else:
                                prefix = match_str[prior_match_span[1]:match_span[0]]
                                tmp_pretty += prefix + click.style(match_str[match_span[0]:match_span[1]], bold=True)
                                prior_match_span = match_span

                        tmp_pretty += match_str[match_span[1]:]
                        pretty.append(attr + ': ' + tag + ': ' + tmp_pretty)

                        counts.append(attr + ': ' + tag + ': ' + str(i+1))

                else:
                    tmp_pretty = ''
                    for i, match in enumerate(hits[doc_id][attr]):
                        result_json[attr] = match.string

                        # pretty handling
                        match_span = match.span()
                        match_str = match.string

                        if i == 0:
                            prefix = match_str[0:match_span[0]]
                            tmp_pretty += prefix + click.style(match_str[match_span[0]:match_span[1]], bold=True)
                            prior_match_span = match_span
                        else:
                            prefix = match_str[prior_match_span[1]:match_span[0]]
                            tmp_pretty += prefix + click.style(match_str[match_span[0]:match_span[1]], bold=True)
                            prior_match_span = match_span

                    tmp_pretty += match_str[match_span[1]:]
                    pretty.append(attr + ': ' + tmp_pretty)

                    counts.append(attr + ': ' + str(i+1))


            if host_url:
                result_json['host_url'] = host_url + doc_id + '.pdf'

            results_json[doc_id] = result_json


            if output == 'counts':
                output_str = click.style(doc_id + '\n', bold=True)

                if host_url:
                    output_str += host_url + doc_id + '.pdf\n'

                attr = 'bib'
                try:
                    output_str += ctx.obj['index'].index[doc_id][attr]['title'] + '\n'
                except:
                    pass

                try:
                    output_str += click.style(ctx.obj['index'].index[doc_id][attr]['author'] + '\n', italic=True)
                except:
                    pass

                output_str += '\n'.join(counts)
                if pager:
                    click.echo_via_pager(output_str)
                else:
                    click.echo(output_str)

            if output == 'pretty':
                output_str = click.style(doc_id + '\n', bold=True, underline=True)

                if host_url:
                    output_str += host_url + doc_id + '.pdf\n'

                attr = 'bib'
                try:
                    output_str += ctx.obj['index'].index[doc_id][attr]['title'] + '\n'
                except:
                    pass

                try:
                    output_str += click.style(ctx.obj['index'].index[doc_id][attr]['author'] + '\n', italic=True)
                except:
                    pass

                output_str += '\n'.join(pretty) + '\n'
                if pager:
                    click.echo_via_pager(output_str)
                else:
                    click.echo(output_str)
                

        if output == 'json':
            if pager:
                click.echo_via_pager(json.dumps(results_json, indent=4))
            else:
                click.echo(json.dumps(results_json, indent=4))


    def print_bib(hits):
        keys = []
        for doc_id in hits:
            attr = 'bib'
            doc_bib = ctx.obj['index'].index[doc_id][attr]

            field = 'key'
            key = doc_bib[field]
            tmp_key = key
            i = 97 # ascii a
            while tmp_key in keys:
                tmp_key = key + chr(i)
                i += 1

            keys.append(tmp_key)
            doc_bib[key] = tmp_key

            click.echo(document.build_bib(doc_bib, bibkey=tmp_key))


    hits = ctx.obj['index'].search(regex)

    if bib:
        print_bib(hits)

    elif url:
        host_url = None
        ssh_config = subprocess.check_output(['ssh', '-G', ctx.obj['config']['database']['host']])
        ssh_config = ssh_config.decode("utf-8").split('\n')

        for line in ssh_config:
            if re.match("hostname", line):
                host_url = line.replace("hostname ", "") + ':' + str(ctx.obj['config']['database']['host_port']) + '/'

        print_results(hits, output=output, pager=pager, host_url=host_url)

    else:
        print_results(hits, output=output, pager=pager)

bibtheque.add_command(search)


#  ──────────────────────────────────────────────────────────────────────────
# print command

@click.command()
@click.argument('doc_id', type=str, default=None) # add bibtex output and handle tags by specifying tag?
@click.pass_context
def cat(ctx, doc_id):
    """Print DOC_ID
    """

    click.echo(json.dumps(ctx.obj['index'].index[doc_id], indent=4))

bibtheque.add_command(cat)


#  ──────────────────────────────────────────────────────────────────────────
# git command

CONTEXT_SETTINGS_git = CONTEXT_SETTINGS.copy()
CONTEXT_SETTINGS_git['ignore_unknown_options'] = True

@click.command('git', short_help='Git manage the index', context_settings=CONTEXT_SETTINGS_git)
@click.argument('git_commands', type=click.UNPROCESSED, nargs=-1)
@click.pass_context
def git(ctx, git_commands):
    """Git command to manage the bibtheque git repository."""

    command = ['git', '-C', ctx.obj['index'].path.parent]

    for arg in git_commands:
        command.append(arg)

    subprocess.run(command)

bibtheque.add_command(git)


#  ──────────────────────────────────────────────────────────────────────────
# cache command

@click.command('cache', short_help='Manage local cache')
@click.option('-c', '--clear', 'clear', is_flag=True, default=False, help='Clear local cache')
@click.option('-s', '--serve', 'serve', is_flag=True, default=False, help='Locally serve the local cache')
@click.pass_context
def cache(ctx, clear, serve):
    """Local cache management."""
    if clear:
        ctx.obj['database'].clear_cache()

    if serve:

        class Handler(http.server.SimpleHTTPRequestHandler):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, directory=ctx.obj['database'].cache, **kwargs)

        handler = Handler

        with socketserver.TCPServer(("", ctx.obj['database'].cache_port), handler) as httpd:
            click.echo('Serving the cache at: locahost:' + str(ctx.obj['database'].cache_port))
            httpd.serve_forever()

bibtheque.add_command(cache)



from click.testing import CliRunner
import subprocess

#  ──────────────────────────────────────────────────────────────────────────
# local imports

from bibtheque.config import config
from bibtheque.cli import bibtheque
from bibtheque.doi import doi_to_bib
from bibtheque.bibtex import bib_to_dict
import bibtheque.database as database
import tests.test_vars as test_vars
import importlib.metadata

#  ──────────────────────────────────────────────────────────────────────────
# running tests

# def test_insert():
    # runner = CliRunner()
    # result = runner.invoke(bibtheque, ['--config', config, 'insert', '--file', 'test.pdf', '-n', 'test notes for test pdf', '--tags', 'test, tags, here', '--dry-run', '--print', test_vars.doi_url])
    # # print(result.output)
    # assert result.exit_code == 0


# def test_search():
    # runner = CliRunner()
    # result = runner.invoke(bibtheque, ['--config', config, 'search', 'notes'])
    # # print(result.output)
    # assert result.exit_code == 0


# def test_search_all():
    # runner = CliRunner()
    # result = runner.invoke(bibtheque, ['--config', config, 'search'])
    # # print(result.output)
    # assert result.exit_code == 0


# def test_regex():
    # runner = CliRunner()
    # result = runner.invoke(bibtheque, ['--config', config, 'regex', '--fields', 'tags', 'here'])
    # # print(result.output)
    # assert result.exit_code == 0


# def test_clean():
    # runner = CliRunner()
    # result = runner.invoke(bibtheque, ['--config', config, 'clean'])
    # assert result.exit_code == 0


# def test_bib():
    # runner = CliRunner()
    # result = runner.invoke(bibtheque, ['--config', config, 'bib', 'test'])

    # result_dict = bib_to_dict(result.output)
    # doi_dict = bib_to_dict(test_vars.doi_bib)

    # del result_dict['key']
    # del doi_dict['key']

    # assert result_dict == doi_dict
    # assert result.exit_code == 0


# def test_write():
    # runner = CliRunner()
    # result = runner.invoke(bibtheque, ['--config', config, 'insert', '--file', 'test.pdf', '-n', 'test notes for test pdf', '--tags', 'test, tags, here', '--force', '--print', test_vars.doi_url])
    # doc_id = result.output.strip('\n')
    # result = runner.invoke(bibtheque, ['--config', config, 'write', doc_id])
    # subprocess.run(['rm', doc_id + '.pdf'])
    # # print(result.output)
    # result = runner.invoke(bibtheque, ['--config', config, 'delete', '--test', doc_id])
    # assert result.exit_code == 0


# def test_modify():
    # runner = CliRunner()
    # result = runner.invoke(bibtheque, ['--config', config, 'insert', '--file', 'test.pdf', '-n', 'test notes for test pdf', '--tags', 'test, tags, here', '--force', '--print', test_vars.doi_url])
    # doc_id = result.output.strip('\n')
    # result = runner.invoke(bibtheque, ['--config', config, 'modify', '--test', doc_id])
    # result = runner.invoke(bibtheque, ['--config', config, 'delete', '--test', doc_id])
    # assert result.exit_code == 0

#  ──────────────────────────────────────────────────────────────────────────
# cleaning test database

# def clean(config):
    # DB = database.Database(config, importlib.metadata.version('bibtheque'))
    # DB.delete()

# clean(config)



from bibtexparser import loads
from bibtexparser.bparser import BibTexParser
from bibtexparser.customization import splitname, convert_to_unicode

# ──────────────────────────────────────────────────────────────────────────
# functions

def bib_to_dict(bib):
    """Returns a dictionary from the given BibTeX citation.

    Parameters
    ----------
    bib : str
        The BibTeX citation string for the desired work.

    Returns
    -------
    bib_dict : dict
        A dictionary containing the BibTeX citation.
    """

    def translator_from_bibtexparser(bib_dict):

        translation_dict = {'ID': 'key', 'ENTRYTYPE': 'type'}
        fields = list(bib_dict.keys())

        for field in fields:
            if field in translation_dict.keys():
                bib_dict[translation_dict[field]] = bib_dict[field]

        for field in fields:
            if field in translation_dict.keys():
                del bib_dict[field]

        return bib_dict


    parser = BibTexParser()
    parser.customization = convert_to_unicode
    bib_dict = loads(bib, parser=parser).entries[0]
    bib_dict = translator_from_bibtexparser(bib_dict)

    return bib_dict


def author_lastname(author_str):
    """Returns author's lastname, given a string containing the author's full name.

    Parameters
    ----------
    author_str : str
        String of the author's name.

    Returns
    -------
    lastname : str
        String of the lastname of the author. 
    """

    name_dict = splitname(author_str)
    lastname = name_dict['last'][0]
    
    return lastname


def build_bib(bib_dict):
    """Generates a BibTeX string from a given dictionary of BibTeX field values.

    Parameters
    ----------
    bib_dict : dict
        Dictionary with all the BibTeX fields.

    Returns
    -------
    bib : str
        BibTeX string of the given dictionary.
    """

    bib = ["@" + bib_dict['type'].lower() + "{" + bib_dict['key'],]

    for field in bib_dict.keys():
        bib += [field + " = " + "{" + bib_dict[field] + "}"]

    bib = ",\n\t".join(bib) + "\n}"

    return bib



#  ──────────────────────────────────────────────────────────────────────────
# local imports

import bibtheque.isbn as isbn
import tests.test_vars as test_vars

# ──────────────────────────────────────────────────────────────────────────
# test functions

def test_is_isbn():
    assert isbn.is_isbn(test_vars.isbn10) == True


def test_get_isbn13():
    assert isbn.get_isbn13(test_vars.isbn10) == test_vars.isbn13


def test_isbn_to_bib():
    assert isbn.isbn_to_bib(test_vars.isbn13) == test_vars.isbn_bib



import numpy as np
from numba import njit

#  ──────────────────────────────────────────────────────────────────────────
# functions

@njit(fastmath=True)
def optimal_string_alignment(a, b):
    """Calculates the Optimal String Alignment distance between two strings.

    Parameters
    ----------
    a : str
        First string to compare.
    b : str
        Second string to compare.

    Returns
    -------
    distance : 
    similarity :

    Notes
    -----
    https://en.wikipedia.org/wiki/Damerau–Levenshtein_distance#Optimal_string_alignment_distance
    """

    M = len(a)
    N = len(b)

    d = np.zeros((M + 1, N + 1))
    
    for i in range(d.shape[0]):
        d[i, 0] = i

    for j in range(d.shape[1]):
        d[0, j] = j

    for i in range(M):
        for j in range(N):

            if a[i] == b[j]:
                cost = 0
            else:
                cost = 1

            I = i + 1
            J = j + 1

            d[I, J] = np.min(np.array([d[I - 1, J] + 1, d[I, J - 1] + 1, d[I - 1, J - 1] + cost]))

            if i > 0 and j > 0 and a[i] == b[j - 1] and a[i - 1] == b[j]:
                d[I, J] = np.min(np.array([d[I, J], d[I - 2, J - 2] + 1]))

    distance = d[-1, -1]
    similarity = (M + N - distance) / (M + N)

    return distance, similarity


def matching(a, b, threshold):
    """
    Determines if strings a and b have a similarity greater than the given threshold.
    """

    _, similarity = optimal_string_alignment(a, b)

    if similarity > threshold:
        return True
    else:
        return False



import isbnlib
from isbnlib.registry import bibformatters
import bibtexparser

# ──────────────────────────────────────────────────────────────────────────
# functions

def is_isbn(url):
    """Check if the given URL has an ISBN

    Parameters
    ----------
    url : str
        URL with potential ISBN.

    Returns
    -------
    bool
        True if there is an ISBN, False if not.
    """

    return not isbnlib.notisbn(url)


def get_isbn13(isbn_gen_str):
    """Returns a clean ISBN number if one exists in the given argument.

    Parameters
    ----------
    isbn_gen_str : str
        Potentially "dirty" string containing an ISBN number, possibly embedded in an URL.

    Returns
    -------
    isbn13 : str
        ISBN 13 number
    """
    
    isbn_gen_str = isbn_gen_str.replace('-', '') # removing all hyphens
    isbn_like = isbnlib.get_isbnlike(isbn_gen_str, level='loose')[-1]  # Get ISBN from a URL or ISBN string.
    clean_isbn_like = isbnlib.clean(isbn_like)

    if isbnlib.is_isbn10(clean_isbn_like):  # Check if the ISBN is ISBN10 and changing it to ISBN13 if it is.
        isbn13 = isbnlib.to_isbn13(clean_isbn_like)
    else:
        isbn13 = clean_isbn_like

    return isbn13


def isbn_to_bib(isbn13):
    """Retrieves the BibTeX citation for a document with a given ISBN-13.

    Parameters
    ----------
    isbn : str
        The ISBN-13 number of the desired document.

    Returns
    -------
    bib : str
        The BibTeX citation of the given ISBN-13.
    """
    
    try:
        isbn_dict = isbnlib.meta(isbn13)
        isbn_bib = bibformatters['bibtex'](isbn_dict)

    except:
        raise

    return isbn_bib
